#
# Copyright 2017-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import print_function

import math

import matplotlib.cbook as cbook
import matplotlib.collections as mcoll
import matplotlib.colors as mcolors
import matplotlib.transforms as mtrans
import matplotlib.transforms as mtransforms
from matplotlib.textpath import TextPath
from matplotlib.patches import PathPatch

import numpy as np

import megaradrp.instrument.constants as cons
import megaradrp.processing.fixrss as fixrss


M_SQRT3 = math.sqrt(3)
M_1_SQRT3 = 1 / M_SQRT3


def hexplot(axis, x, y, z, scale=1.0, extent=None,
            cmap=None, norm=None, vmin=None, vmax=None,
            alpha=None, linewidths=None, edgecolors='none',
            **kwargs):
    """
    Make a hexagonal grid plot.

    Returns
    =======
    object: matplotlib.collections.PolyCollection

    """
    axis._process_unit_info(xdata=x, ydata=y, kwargs=kwargs)

    x, y, z = cbook.delete_masked_points(x, y, z)

    x = np.array(x, float)
    y = np.array(y, float)

    sx = 2 * M_1_SQRT3 * scale * 0.99
    sy = scale * 0.99

    if extent is not None:
        xmin, xmax, ymin, ymax = extent
    else:
        xmin, xmax = (np.amin(x - sx), np.amax(x + sx)) if len(x) else (0, 1)
        ymin, ymax = (np.amin(y - sy), np.amax(y + sy)) if len(y) else (0, 1)

        # to avoid issues with singular data, expand the min/max pairs
        xmin, xmax = mtrans.nonsingular(xmin, xmax, expander=0.1)
        ymin, ymax = mtrans.nonsingular(ymin, ymax, expander=0.1)

    padding = 1.0e-9 * (xmax - xmin)
    xmin -= padding
    xmax += padding

    n = len(x)
    polygon = np.zeros((6, 2), float)

    mx = my = 0.99 * scale
    polygon[:, 0] = mx * np.array([-0.5 * M_1_SQRT3, 0.5 * M_1_SQRT3, 1.0 * M_1_SQRT3,
                                   0.5 * M_1_SQRT3, -0.5 * M_1_SQRT3, -1.0 * M_1_SQRT3])
    polygon[:, 1] = my * np.array([0.5, 0.5, 0.0, -0.5, -0.5, 0.0])

    offsets = np.zeros((n, 2), float)
    offsets[:, 0] = x
    offsets[:, 1] = y

    collection = mcoll.PolyCollection(
        [polygon],
        edgecolors=edgecolors,
        linewidths=linewidths,
        offsets=offsets,
        transOffset=mtransforms.IdentityTransform(),
        offset_position="data"
    )

    if isinstance(norm, mcolors.LogNorm):
        if (z == 0).any():
            # make sure we have not zeros
            z += 1

    if norm is not None:
        if norm.vmin is None and norm.vmax is None:
            norm.autoscale(z)

    if norm is not None and not isinstance(norm, mcolors.Normalize):
        msg = "'norm' must be an instance of 'mcolors.Normalize'"
        raise ValueError(msg)

    collection.set_array(z)
    collection.set_cmap(cmap)
    collection.set_norm(norm)
    collection.set_alpha(alpha)
    collection.update(kwargs)

    if vmin is not None or vmax is not None:
        collection.set_clim(vmin, vmax)
    else:
        collection.autoscale_None()

    corners = ((xmin, ymin), (xmax, ymax))
    axis.update_datalim(corners)
    axis.autoscale_view(tight=True)

    # add the collection last
    axis.add_collection(collection, autolim=False)
    return collection


def transform_to_pixel(wcs, coords, ref=0):
    # We need only the WL axis
    coords_m = [coords, np.zeros_like(coords)]
    r = wcs.wcs_world2pix(np.transpose(coords_m), ref)
    first_col = r[:, 0].astype('int')
    return first_col


def is_inside(shape, *coords):

    for coord in coords:
        for shape1d, crds in zip(shape, coord):
            if not is_inside_1d(shape1d, crds):
                return False
    return True


def is_inside_1d(shape1d, coords):
    for coord in coords:
        if coord < 0:
            return False
        if coord >= shape1d:
            return False
    return True


def extract_region(wcs_wl, region, rssdata):
    cut1, cut2 = transform_to_pixel(wcs_wl, region)

    if is_inside_1d(rssdata.shape[1], [cut1, cut2]):
        z = rssdata[:, cut1:cut2].mean(axis=1)
        return z
    else:
        raise ValueError('region to average is outside image')


def extract_column(wcs_wl, col, rssdata):
    cut, = transform_to_pixel(wcs_wl, [col])

    if is_inside_1d(rssdata.shape[1], [cut]):
        # cut1, cut2 = args.average_region
        z = rssdata[:, cut]
        return z
    else:
        raise ValueError('column is outside image')


def extract_zval(rssdata, wcs_wl, coordinate_type, column=None, average_region=None, continuum_region=None):

    if continuum_region is None and average_region is None:
        raise ValueError('continuum_region and average_region are None')

    if rssdata.ndim == 1:
        # data is 1d
        print('RSS is 1D')
        zval = rssdata
    elif rssdata.ndim == 2:
        print('world coordinates in', coordinate_type)
        if column is not None:
            print('use column', column)
            zval = extract_column(wcs_wl, column, rssdata)
        else:
            print('compute average in', average_region)
            zval = extract_region(wcs_wl, average_region, rssdata)
        # Subtract continuum
        if continuum_region:
            print('compute continuum in', continuum_region)
            z0 = extract_region(wcs_wl, continuum_region, rssdata)
            zval -= z0
    else:
        raise ValueError('image must be 1D or 2D')
    return zval


def main(argv=None):
    """Process command line"""

    import argparse
    import json

    import matplotlib.pyplot as plt
    import astropy.io.fits as fits
    from astropy.wcs import WCS
    from astropy.visualization import simple_norm
    import astropy.units as u

    import megaradrp.datamodel as dm
    from megaradrp.instrument.focalplane import FocalPlaneConf
    from megaradrp.processing.cube import create_cube_from_rss
    from megaradrp.processing.wcs import update_wcs_from_ipa, compute_pa_from_ipa

    # scale of the LCB grid in mm
    scale_lcb = cons.SPAXEL_SCALE.to(u.mm).value

    parser = argparse.ArgumentParser(description='Display MEGARA RSS images')
    parser.add_argument('--wcs-grid', '--display-wcs-grid', action='store_true',
                        help='Display WCS grid')
    parser.add_argument('--wcs-pa-from-header', action='store_true',
                        help="Use PA angle from PC keys", dest='pa_from_header')
    parser.add_argument('--fix-missing', action='store_true',
                        help="Interpolate missing fibers")
    parser.add_argument('--average-region', nargs=2, default=[1000, 3000],
                        type=int, help='Region of the RSS averaged on display')
    parser.add_argument('--extname', '-e', default='PRIMARY',
                        help='Name of the extension used')
    parser.add_argument('--column', '-c', type=int,
                        help='Column of the RSS on display')
    parser.add_argument('--continuum-region', nargs=2,
                        type=int,
                        help='Region of the RSS used for continuum subtraction')
    parser.add_argument('--coordinate-type', choices=['pixel', 'wcs'],
                        default='pixel',
                        help='Types of coordinates used')
    parser.add_argument('--colormap', type=plt.get_cmap,
                        help='Name of a valid matplotlib colormap')
    parser.add_argument('--plot-sky', '--display-sky', action='store_true',
                        help='Display SKY bundles')
    parser.add_argument('--display-fibid', '--plot-fibid', action='store_true',
                        help='Display fiber IDs of the spaxels')
    parser.add_argument('--plot-nominal-config', action='store_true',
                        help='Plot nominal configuration, do not use the header')
    parser.add_argument('--hide-values', action='store_true',
                        help='Do not show values out of range')
    parser.add_argument('--title', help='Title of the plot')
    parser.add_argument('--label', help='Legend of the colorbar')
    parser.add_argument('--hex-size', type=float,
                        help=f'Size of the hexagons (default is {scale_lcb})',
                        default=scale_lcb)
    parser.add_argument('--hex-rel-size', type=float,
                        help='Scale the size of hexagons by a factor',
                        default=1.0)
    parser.add_argument('--min-cut', type=float,
                        help='Inferior cut level')
    parser.add_argument('--max-cut', type=float,
                        help='Superior cut level')
    parser.add_argument('--percent', type=float,
                        help='Compute cuts using percentiles')
    parser.add_argument('--stretch',
                        choices=['linear', 'sqrt', 'power', 'log', 'asinh'],
                        default='linear',
                        help='Name of the strech method used for display'
                        )

    group_c = parser.add_argument_group('contouring')
    group_c.add_argument('--contour-pixel-size', type=float, default=0.4,
                         help="Pixel size in arc seconds for image reconstruction")
    group_c.add_argument('--contour-levels',
                         help="Contour levels")
    group_c.add_argument('--contour', action='store_true',
                         help="Draw contours")
    group_c.add_argument('--contour-image',
                         help="Image for computing contours")
    group_c.add_argument('--contour-image-column', type=int,
                         help='Column of image used for contouring')
    group_c.add_argument('--contour-image-save',
                         help='Save image used for contouring')
    group_c.add_argument('--contour-image-region', nargs=2, default=[1000, 3000],
                         type=int, help='Region of the image used for contouring')
    group_c.add_argument('--contour-is-density', action='store_true',
                         help='The data is a magnitude that does not require scaling')

    parser.add_argument('rss', metavar='RSS', nargs='+',
                        help='RSS images to process')

    args = parser.parse_args(args=argv)

    for fname in args.rss:
        with fits.open(fname) as img:
            if args.fix_missing:
                fibid = 623
                print(f'interpolate fiber {fibid}')
                img = fixrss.fix_missing_fiber(img, fibid)
            if args.plot_nominal_config:
                insmode = img['FIBERS'].header['INSMODE']
                fp_conf = dm.get_fiberconf_default(insmode)
            else:
                fp_conf = FocalPlaneConf.from_img(img)
            plot_mask = np.ones((fp_conf.nfibers,), dtype=np.bool)
            if not args.plot_sky:
                skyfibers = fp_conf.sky_fibers()
                skyfibers.sort()
                skyfibers_idx = [(fibid - 1) for fibid in skyfibers]
                plot_mask[skyfibers_idx] = False

            x0 = np.empty((fp_conf.nfibers,))
            y0 = np.empty((fp_conf.nfibers,))
            num = {}
            names = {}
            # Key is fibid
            for _, fiber in sorted(fp_conf.fibers.items()):
                idx = fiber.fibid - 1
                x0[idx] = fiber.x
                y0[idx] = fiber.y
                num[idx] = fiber.fibid
                names[idx] = fiber.name

            extname = args.extname
            if args.coordinate_type == 'wcs':
                # read WCS from extname
                wcs_wl = WCS(img[extname].header)
            else:
                wcs_wl = WCS(naxis=len(img[extname].shape))

            rssdata = np.squeeze(img[extname].data)

            zval0 = extract_zval(
                rssdata, wcs_wl, args.coordinate_type,
                args.column, args.average_region, args.continuum_region
            )
            fig = plt.figure()
            
            x = x0[plot_mask]
            y = y0[plot_mask]
            zval = zval0[plot_mask]
            projection = None

            if args.wcs_grid:
                print('display wcs grid')
                if not args.pa_from_header:
                    print('compute PA from IPA')
                    ipa = img[extname].header['IPA']
                    pa = compute_pa_from_ipa(ipa)
                    print('IPA is', ipa, ' PA is', pa)
                    hdr_fibers = update_wcs_from_ipa(img['FIBERS'].header, pa)
                else:
                    hdr_fibers = img['FIBERS'].header

                projection = WCS(hdr_fibers)

            # plt.subplots_adjust(hspace=0.5)
            ax = fig.add_axes([0.15, 0.1, 0.8, 0.8], projection=projection)

#            ax.set_xlim([-6.5, 6.5])
#            ax.set_ylim([-6, 6])

            if args.wcs_grid:
                ax.coords.grid(color='black', alpha=1.0, linestyle='solid')

            norm = simple_norm(zval, args.stretch,
                               min_cut=args.min_cut,
                               max_cut=args.max_cut,
                               percent=args.percent
                               )

            if args.hide_values:
                zdisp = zval.copy()
                if norm.vmin is not None:
                    zdisp[zval < norm.vmin] = np.nan
                if norm.vmax is not None:
                    zdisp[zval > norm.vmax] = np.nan
            else:
                zdisp = zval
            scale = args.hex_rel_size * args.hex_size
            col = hexplot(ax, x, y, zdisp, scale=scale, cmap=args.colormap, norm=norm)
            if args.display_fibid:
                # Plot spaxel labels
                for xx, yy, fid in zip(x, y, num):
                    tp1 = TextPath(
                        (xx-0.05, yy+0.16), f"{fid:03d}", size=0.045)
                    tp2 = TextPath(
                        (xx-0.05, yy-0.18), f"{names[fid]}", size=0.045)
                    ax.add_patch(PathPatch(tp1, zorder=2, lw=0, fc="black"))
                    ax.add_patch(PathPatch(tp2, zorder=2, lw=0, fc="black"))

            if args.title is not None:
                ax.set_title(args.title)

            cb = fig.colorbar(col)
            if args.label is not None:
                cb.set_label(args.label)

            if args.contour:
                target_scale_arcsec = args.contour_pixel_size

                if args.contour_image is not None:
                    with fits.open(args.contour_image) as hdu_con:
                        extname = args.extname

                        if args.coordinate_type == 'wcs':
                            # read WCS from extname
                            wcs_wl_c = WCS(hdu_con[extname].header)
                        else:
                            wcs_wl_c = WCS(naxis=len(hdu_con[extname].shape))

                        rssdata_con = np.squeeze(hdu_con[extname].data)

                        zval_c = extract_zval(rssdata_con,
                                              wcs_wl_c,
                                              args.coordinate_type,
                                              column=args.contour_image_column,
                                              average_region=args.contour_image_region,
                                              continuum_region=None)
                else:
                    zval_c = zval0

                # Build synthetic rss... for reconstruction
                primary = fits.PrimaryHDU(data=zval_c[:, np.newaxis], header=img[extname].header)
                synt = fits.HDUList([primary, img['FIBERS']])

                if args.contour_image_save is not None:
                    synt.writeto(args.contour_image_save)

                conserve_flux = not args.contour_is_density
                order = 1
                s_cube = create_cube_from_rss(synt, order, target_scale_arcsec, conserve_flux=conserve_flux)
                cube_wcs = WCS(s_cube[0].header).celestial
                px, py = cube_wcs.wcs.crpix
                interp = np.squeeze(s_cube[0].data)
                td = mtransforms.Affine2D().translate(-px, -py).scale(target_scale_arcsec, target_scale_arcsec)
                tt_d = td + ax.transData
                # im = ax.imshow(interp, alpha=0.9, cmap='jet', transform=tt_d)
                # im = ax.imshow(interp, alpha=0.9, cmap='jet', transform=ax.get_transform(cube_wcs))
                if args.contour_levels is not None:
                    levels = json.loads(args.contour_levels)
                    mm = ax.contour(interp, levels, transform=tt_d)
                else:
                    mm = ax.contour(interp, transform=tt_d)
                print('contour levels', mm.levels)

            plt.show()


if __name__ == '__main__':
    main()



#
# Copyright 2016-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Data model for MEGARA"""

from __future__ import division


import pkgutil
import logging

import astropy.io.fits as fits
from io import StringIO
from numina.datamodel import DataModel, QueryAttribute, KeyDefinition
from numina.util.convert import convert_date

from megaradrp.datatype import MegaraDataType, DataOrigin
import megaradrp.instrument.constants as cons


_logger = logging.getLogger(__name__)


class MegaraDataModel(DataModel):
    """Data model of MEGARA images"""

    query_attrs = {
        'vph': QueryAttribute('vph', str),
        'insmode': QueryAttribute('insmode', str),
        'insconf': QueryAttribute('insconf', str),
        'speclamp': QueryAttribute('speclamp', str),
        'temp': QueryAttribute('temp', float),
        'confid': QueryAttribute('confid', str)
    }

    meta_info_headers = [
        'instrument',
        'object',
        'observation_date',
        'uuid',
        'type',
        'mode',
        'exptime',
        'darktime',
        'insconf',
        'blckuuid',
        'quality_control',
        'vph',
        'insmode'
    ]

    db_info_keys = [
        'instrument',
        'object',
        'observation_date',
        'uuid',
        'type',
        'mode',
        'exptime',
        'darktime',
        'insconf',
        'blckuuid',
        'quality_control',
        'vph',
        'insmode'
    ]

    db_info_keys_extra = [
        'vph',
        'insmode'
    ]

    meta_dinfo_headers = [
        'exptime',
        'observation_date',
        'vph',
        'vphpos',
        'insmode',
        'focus',
        'osfilter',
        'uuid',
        'temp',
        'block_uuid',
        'insconf_uuid',
        'speclamp',
        'imgid'
    ]

    PLATESCALE = cons.GTC_FC_A_PLATESCALE.value

    def __init__(self):

        instrument_mappings = {
            'date_obs': ('DATE-OBS', 0, convert_date),
            'insconf': 'insconf',
            'insconf_uuid': 'insconf',
            'blckuuid': 'blckuuid',
            'block_uuid': 'blckuuid',
            'vph': ('VPH', 'undefined'),
            'vphpos': ('VPHWHPOS', 'undefined'),
            'focus': ('FOCUS', 'undefined'),
            'osfilter': ('OSFILTER', 'undefined'),
            'temp': ('SENTEMP4', 0.0),
            'speclamp': ('SPECLAMP', 'undefined'),
            'confid': KeyDefinition('CONFID', ext='FIBERS'),
        }
        super(MegaraDataModel, self).__init__(
            'MEGARA',
            instrument_mappings
        )

    def get_fiberconf(self, img):
        """Obtain FiberConf from image"""
        return get_fiberconf(img)

    def gather_info_oresult(self, val):
        return [self.gather_info_dframe(f) for f in val.images]

    def fiber_scale_unit(self, img, unit=False):
        return fiber_scale_unit(img, unit=unit)


def fiber_scale_unit(img, unit=False):
    funit = img['FIBERS'].header.get("FUNIT", "arcsec")

    if funit == "arcsec":
        scale = 1
    else:
        scale = cons.GTC_FC_A_PLATESCALE.value
    if unit:
        return scale, funit
    else:
        return scale


def get_fiberconf(img):
    """Obtain FiberConf from image"""

    main_insmode = img[0].header.get('INSMODE', '')

    if 'FIBERS' in img:
        # We have a 'fibers' extension
        # Information os there
        hdr_fiber = img['FIBERS'].header
        return read_fibers_extension(hdr_fiber, insmode=main_insmode)
    else:
        return get_fiberconf_default(main_insmode)


def create_default_fiber_header(insmode):
    """Obtain default FIBER header"""
    if insmode == 'LCB':
        slit_file = 'lcb_default_header.txt'
    elif insmode == 'MOS':
        slit_file = 'mos_default_header.txt'
    else:
        # Read fiber info from headers
        raise ValueError(f'Invalid INSMODE {insmode}')

    data = pkgutil.get_data('megaradrp.instrument.configs', slit_file)
    default_hdr = StringIO(data.decode('utf8'))
    hdr_fiber = fits.header.Header.fromfile(default_hdr)
    return hdr_fiber


def get_fiberconf_default(insmode):
    """Obtain default FiberConf object"""
    hdr_fiber = create_default_fiber_header(insmode)
    return read_fibers_extension(hdr_fiber)


def read_fibers_extension(hdr, insmode='LCB'):
    """Read the FIBERS extension

    Parameters
    ==========
    hdr :
       FITS header
    insmode : str
        default INSMODE

    Returns
    =======
    FibersConf


    """
    import megaradrp.instrument.focalplane as fp
    return fp.FocalPlaneConf.from_header(hdr)


def describe_hdulist_megara(hdulist):
    prim = hdulist[0].header
    instrument = prim.get("INSTRUME", "unknown")
    image_type = prim.get("IMAGETYP")
    if image_type is None:
        # try this also
        image_type = prim.get("NUMTYPE")

    # date_obs = convert_date(prim.get("DATE-OBS"))
    date_obs = prim.get("DATE-OBS")
    img_uuid = prim.get("UUID")
    insconf = prim.get("INSCONF", 'undefined')

    if image_type is None:
        # inferr from header
        datatype = megara_inferr_datetype_from_image(hdulist)
    else:
        datatype = MegaraDataType[image_type]

    obs = {}
    proc = {}
    if datatype.value < MegaraDataType.IMAGE_PROCESSED.value:
        origin = DataOrigin.OBSERVED
    else:
        origin = DataOrigin.PROCESSED

    return {'instrument': instrument, 'datatype': datatype,
            'origin': origin, 'uuid': img_uuid,
            'insconf': insconf,
            'observation': obs,
            'observation_date': date_obs,
            'processing': proc
            }


def megara_inferr_datatype(obj):

    if isinstance(obj, fits.HDUList):
        return megara_inferr_datetype_from_image(obj)
    elif isinstance(obj, dict):
        return megara_inferr_datetype_from_dict(obj)
    else:
        raise TypeError(f"I don't know how to inferr datatype from {obj}")


def megara_inferr_datetype_from_dict(obj):
    # this comes from JSON
    dtype = obj['type_fqn']
    if dtype in ["megaradrp.products.tracemap.TraceMap"]:
        return MegaraDataType.TRACE_MAP
    elif dtype in ["megaradrp.products.modelmap.ModelMap"]:
        return MegaraDataType.MODEL_MAP
    elif dtype in ["megaradrp.products.wavecalibration.WavelengthCalibration"]:
        return MegaraDataType.WAVE_CALIB
    else:
        return MegaraDataType.UNKNOWN


def megara_inferr_datetype_from_image(hdulist):
    IMAGE_RAW_SHAPE = (4212, 4196)
    IMAGE_PROC_SHAPE = (4112, 4096)
    RSS_IFU_PROC_SHAPE = (623, 4096)
    RSS_MOS_PROC_SHAPE = (644, 4096)
    RSS_IFU_PROC_WL_SHAPE = (623, 4300)
    RSS_MOS_PROC_WL_SHAPE = (644, 4300)
    SPECTRUM_PROC_SHAPE = (4300,)
    prim = hdulist[0].header

    image_type = prim.get("IMAGETYP")
    if image_type is None:
        # try this also
        image_type = prim.get("NUMTYPE")

    if image_type is not None:
        datatype = MegaraDataType[image_type]
        return datatype

    pshape = hdulist[0].shape
    obsmode = prim.get("OBSMODE", "unknown")
    if pshape == IMAGE_RAW_SHAPE:
        datatype = MegaraDataType.IMAGE_RAW
    elif pshape == IMAGE_PROC_SHAPE:
        datatype = MegaraDataType.IMAGE_PROCESSED
    elif pshape == RSS_IFU_PROC_SHAPE: # IFU
        datatype = MegaraDataType.RSS_PROCESSED
    elif pshape == RSS_MOS_PROC_SHAPE: # MOS
        datatype = MegaraDataType.RSS_PROCESSED
    elif pshape == RSS_IFU_PROC_WL_SHAPE: # IFU
        datatype = MegaraDataType.RSS_WL_PROCESSED
    elif pshape == RSS_MOS_PROC_WL_SHAPE: # MOS
        datatype = MegaraDataType.RSS_WL_PROCESSED
    elif pshape == SPECTRUM_PROC_SHAPE:
        datatype = MegaraDataType.SPEC_PROCESSED
    else:
        datatype = MegaraDataType.UNKNOWN

    if datatype == MegaraDataType.IMAGE_RAW:
        if obsmode in ["MegaraSuccess", "MegaraFail"]:
            sub_datatype = MegaraDataType.IMAGE_TEST
        elif obsmode in ["MegaraBiasImage"]:
            sub_datatype = MegaraDataType.IMAGE_BIAS
        elif obsmode in ["MegaraDarkImage"]:
            sub_datatype = MegaraDataType.IMAGE_DARK
        elif obsmode in ["MegaraSlitFlat"]:
            sub_datatype = MegaraDataType.IMAGE_SLITFLAT
        elif obsmode in ["MegaraFiberFlatImage", "MegaraTraceMap", "MegaraModelMap"]:
            sub_datatype = MegaraDataType.IMAGE_FLAT
        elif obsmode in ["MegaraArcCalibration"]:
            sub_datatype = MegaraDataType.IMAGE_COMP
        elif obsmode in ["MegaraTwilightFlatImage"]:
            sub_datatype = MegaraDataType.IMAGE_TWILIGHT
        elif obsmode in ["MegaraLcbImage", "MegaraMosImage"]:
            sub_datatype = MegaraDataType.IMAGE_TARGET
        elif obsmode in ["MegaraMosStdStar", "MegaraExtinctionStar",
                         "MegaraLcbStdStar", "MegaraSensitivityStar"]:
            sub_datatype = MegaraDataType.IMAGE_TARGET
        elif obsmode in ["MegaraFocusTelescope",
                         "MegaraLcbAcquisition", "MegaraMosAcquisition"]:
            sub_datatype = MegaraDataType.IMAGE_TARGET
        elif obsmode in ["MegaraBadPixelMask", "MegaraFocusSpectrograph"]:
            sub_datatype = MegaraDataType.IMAGE_RAW
        else:
            sub_datatype = MegaraDataType.UNKNOWN

        return sub_datatype
    elif datatype == MegaraDataType.SPEC_PROCESSED:
        numrnam = prim.get("NUMRNAM", "unknown")
        if numrnam in ['LCBStandardRecipe', 'MOSStandardRecipe']:
            sub_datatype = MegaraDataType.MASTER_SENSITIVITY
        else:
            sub_datatype = datatype
        return sub_datatype
    return datatype


def check_obj_megara(obj, astype=None, level=None):
    import megaradrp.validators as val
    if astype is None:
        datatype = megara_inferr_datatype(obj)
        _logger.debug(f'check object as it says it is ({datatype})')
        thistype = datatype
    else:
        _logger.debug(f'check object as {astype}')
        thistype = astype
    checker = val.check_as_datatype(thistype)
    res = checker(obj, level=level)
    return res



#
# Copyright 2016-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Focal plane description for MEGARA"""

from __future__ import division


import math
import re
import warnings

import astropy.coordinates

from .ienums import TargetType, BundleType
from megaradrp.processing.hexgrid import connected6


class FocalPlaneConf(object):
    """Configuration of focal plane"""
    def __init__(self, name='LCB'):
        self.name = name
        self.conf_id = "00000000-0000-0000-0000-000000000000"
        bundles = dict()
        if name == 'LCB':
            # 1 LCB + 8 SKY
            bundles[0] = LcbBundleConf()
            for i in range(93, 100 + 1):
                bundles[i] = SkyBundleConf(i)

        elif name == 'MOS':
            # 92 RPs
            for i in range(1, 92 + 1):
                bundles[i] = BundleConf(i, BundleType.RP)
        else:
            raise ValueError(f"name {name} is invalid")
        self.nbundles = len(bundles)
        self.nfibers = sum(bundle.nfibers for bundle in bundles.values())
        self.bundles = bundles
        self.fibers = {}
        self.funit = "mm"

    def attach_fibers(self, fibers):
        bun_fib = {}
        # Fibers in each bundle
        for fibid, ff in fibers.items():
            if ff.bundle_id not in bun_fib:
                bun_fib[ff.bundle_id] = []
            bun_fib[ff.bundle_id].append(ff.fibid)

        for bid, bundle in self.bundles.items():
            # Attach the fibers
            bun_fibers = {fibid: fibers[fibid] for fibid in bun_fib[bid]}
            bundle.attach_fibers(bun_fibers)
        self.fibers = fibers

    @classmethod
    def from_header(cls, hdr):
        """Create a FocalPlaneConf object from map-like header"""

        # defaults = dict()
        # defaults['LCB'] = (9, 623)
        # defaults['MOS'] = (92, 644)

        insmode = hdr.get('INSMODE')
        confid = hdr.get('CONFID', "00000000-0000-0000-0000-000000000000")

        conf = FocalPlaneConf(insmode)
        conf.conf_id = confid

        if conf.nbundles != hdr.get('NBUNDLES'):
            raise ValueError(f'checking NBUNDLES != {conf.nbundles}')
        if conf.nfibers != hdr.get('NFIBERS'):
            raise ValueError(f'checking NFIBERS != {conf.nfibers}')

        conf.funit = hdr.get("FUNIT", "arcsec")
        # Read bundles

        bun_ids = []
        fib_ids = []
        fibers = {}

        # loop over everything, count BUN%03d_P and FIB%03d_B
        pattern1 = re.compile(r"BUN(\d+)_P")
        pattern2 = re.compile(r"FIB(\d+)_B")
        for key in hdr:
            bun_re = pattern1.match(key)
            fib_re = pattern2.match(key)
            if bun_re:
                bun_idx = int(bun_re.group(1))
                bun_ids.append(bun_idx)
            elif fib_re:
                fib_idx = int(fib_re.group(1))
                fib_ids.append(fib_idx)

        for fibid in fib_ids:
            ff = FiberConf.from_header(hdr, fibid)
            # if ff.bundle_id not in bun_fib:
            #     bun_fib[ff.bundle_id] = []
            # bun_fib[ff.bundle_id].append(ff.fibid)
            fibers[ff.fibid] = ff

        for bid in bun_ids:
            # Get bundle
            bb = conf.bundles[bid]
            bb.target_priority = hdr["BUN%03d_P" % bid]
            bb.target_name = hdr["BUN%03d_I" % bid]
            bb.target_type = TargetType[hdr["BUN%03d_T" % bid]]
            bb.enabled = hdr.get("BUN%03d_E" % bid, True)
            bb.x = hdr.get("BUN%03d_X" % bid, 0.0)
            bb.y = hdr.get("BUN%03d_Y" % bid, 0.0)
            bb.pa = hdr.get("BUN%03d_O" % bid, 0.0)

        conf.attach_fibers(fibers)

        # Double check
        if conf.name == 'LCB':
            refid = 614
            ref_fiber = conf.fibers[refid]
            if ref_fiber.x < -6:
                # arcsec
                if conf.funit != 'arcsec':
                    print('warning')
            else:
                # mm
                if conf.funit != 'mm':
                    print('warning')
        return conf

    @classmethod
    def from_img(cls, img):
        """Create a FocalPlaneConf object from a FITS image"""
        return cls.from_header(img['FIBERS'].header)

    def sky_fibers(self, valid_only=False, ignored_bundles=None):
        result = []
        if ignored_bundles is None:
            ignored_bundles = []

        for bundle in self.bundles.values():
            if bundle.id in ignored_bundles:
                continue
            if bundle.target_type is TargetType.SKY:
                if valid_only:
                    for fib in bundle.fibers.values():
                        if fib.valid:
                            result.append(fib.fibid)
                else:
                    result.extend(bundle.fibers.keys())
        return result

    def connected_fibers(self, valid_only=False):
        """Return the fibers connected in the IFU"""
        if self.name == 'MOS':
            return []

        result = []
        for bundle in self.bundles.values():
            if bundle.target_type is not TargetType.SKY:
                if valid_only:
                    for fib in bundle.fibers.values():
                        if fib.valid:
                            result.append(fib)
                else:
                    result.extend(bundle.fibers.values())
        return result

    def nearby_fibers(self, fibid):
        """Obtain the fiber IDs of fibers around fibid"""

        # Find the bundle the fiber belongs to
        try:
            fiber_conf = self.fibers[fibid]
            bundle = self.bundles[fiber_conf.bundle_id]
            result = bundle.nearby(fibid)
            return result
        except KeyError as err:
            # equivalent to
            # raise ValueError from None
            ex = ValueError(f'fibid {fibid} does not exist')
            raise ex from err

    def inactive_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if fiber.inactive:
                result.append(fiber.fibid)
        return result

    def active_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if not fiber.inactive:
                result.append(fiber.fibid)
        return result

    def valid_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if fiber.valid:
                result.append(fiber.fibid)
        return result

    def invalid_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if not fiber.valid:
                result.append(fiber.fibid)
        return result

    def spectral_coverage(self):
        lowc = []
        upperc = []
        for fibid, r in self.fibers.items():
            if r.w1:
                lowc.append(r.w1)
            if r.w2:
                upperc.append(r.w2)

        mn = max(lowc)
        nn = min(lowc)

        mx = min(upperc)
        nx = max(upperc)
        return (mn, mx), (nn, nx)

    def bundles_to_table(self):
        """Convert bundles to a Table"""

        import astropy.table

        attrnames = ['id', 'x', 'y', 'pa', 'enabled',
                     'target_type', 'target_priority', 'target_name']
        cnames = ['bundle_id', 'x', 'y', 'pa', 'enabled',
                  'target_type', 'target_priority', 'target_name']
        obj_data = {}
        for a, c in zip(attrnames, cnames):
            obj_data[c] = [getattr(ob, a) for ob in self.bundles.values()]
        result = astropy.table.Table(obj_data, names=cnames)
        result['x'].unit = self.funit
        result['y'].unit = self.funit
        result['pa'].unit = 'deg'
        return result

    def fibers_to_table(self):
        """Convert fibers to a Table"""
        import astropy.table
        attrnames = ['fibid', 'name', 'x', 'y', 'inactive', 'valid',
                     'bundle_id']
        cnames = ['fibid', 'name', 'x', 'y', 'inactive', 'valid',
                  'bundle_id']
        obj_data = {}

        for a, c in zip(attrnames, cnames):
            obj_data[c] = [getattr(ob, a) for ob in self.fibers.values()]
        result = astropy.table.Table(obj_data, names=cnames)
        result['x'].unit = self.funit
        result['y'].unit = self.funit
        return result


class FiberConfs(FocalPlaneConf):
    """Configuration of focal plane

    .. deprecated:: 0.10
            `FiberConfs` is replaced by `FocalPlaneConf`. It will
            be removed in 1.0

    """
    def __init__(self):
        super(FiberConfs, self).__init__()
        warnings.warn("The 'FiberConfs' class was renamed to 'FocalPlaneConf'", DeprecationWarning, stacklevel=2)


class BundleConf(object):
    """Description of a bundle"""
    def __init__(self, bundle_id, bundle_type, target_type=TargetType.UNASSIGNED,
                 target_priority=0, target_name='unknown', enabled=True, movable=True):
        self.id = bundle_id
        self.name = "unknown"
        self.bundle_type = bundle_type
        self.target_type = target_type
        self.target_priority = target_priority
        self.target_name = target_name
        self.enabled = enabled
        self.movable = movable
        self.nfibers = 7
        if self.bundle_type == BundleType.LCB:
            self.nfibers = 567
        self.x_fix = 0
        self.y_fix = 0
        self.pa_fix = 0
        self.x = 0
        self.y = 0
        self.pa = 0
        # Experimental
        self.fibers = {}
        self._map1 = {}
        self._map2 = {}

    def attach_fibers(self, fibers):
        self.fibers = fibers
        # Experimental
        cos_30 = math.sqrt(3) / 2
        # This is ~= SPAXEL_SCALE
        dd = 0.44225029050703585
        self._map1 = {}
        self._map2 = {}
        for r in self.fibers.values():
            ii = r.x / (dd * cos_30)
            jj = (r.y / dd - 0.5) - ii * 0.5
            uv = (round(ii), round(jj))
            self._map1[r.fibid] = uv
            self._map2[uv] = r.fibid

    def nearby(self, fibid):
        """Find fiber ids around a central fibid"""
        uv1 = self._map1[fibid]
        result = []
        for con in connected6(*uv1):
            try:
                result.append(self._map2[con])
            except KeyError:
                pass
        return result


class LcbBundleConf(BundleConf):
    """Description of the LCB bundle"""
    def __init__(self, bundle_id=0):
        super(LcbBundleConf, self).__init__(
            bundle_id=bundle_id, bundle_type=BundleType.LCB,
            movable=False
        )
        self.name = "LCB"
        self.nrows = 21
        self.ncols = 27


class SkyBundleConf(BundleConf):
    """Description of a sky bundle"""
    def __init__(self, bundle_id):
        super(SkyBundleConf, self).__init__(
            bundle_id=bundle_id, bundle_type=BundleType.SKY,
            target_type=TargetType.SKY, movable=False
        )


class FiberConf(object):
    """Description of the fiber"""
    def __init__(self, fibid=0, bundle_id=None, inactive=False):
        self.fibid = fibid
        self.name = 'unknown'
        self.bundle_id = bundle_id
        self.inactive = inactive
        self.valid = True
        self.x = 0.0
        self.y = 0.0
        self.coord = None

    @classmethod
    def from_header(cls, hdr, fibid):
        ff = FiberConf(fibid=fibid)
        # Coordinates
        dec = hdr["FIB%03d_D" % fibid]
        ra = hdr["FIB%03d_R" % fibid]
        frame = 'icrs'
        ff.d = dec
        ff.r = ra
        ff.coord = astropy.coordinates.SkyCoord(
            ra, dec, frame=frame, unit='deg'
        )
        ff.o = 0  # hdr["FIB%03d_O" % fibid]
        # Active
        ff.inactive = not hdr["FIB%03d_A" % fibid]

        # Coordinates XY
        ff.x = hdr["FIB%03d_X" % fibid]
        ff.y = hdr["FIB%03d_Y" % fibid]

        ff.bundle_id = hdr["FIB%03d_B" % fibid]
        ff.name = hdr.get("FIB%03d_N" % fibid, 'unknown')

        ff.w1 = hdr.get("FIB%03dW1" % fibid, None)
        ff.w2 = hdr.get("FIB%03dW2" % fibid, None)

        # Validity
        if ff.inactive:
            ff.valid = False
        else:
            ff.valid = hdr.get("FIB%03d_V" % fibid, True)
        return ff



#
# Copyright 2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""MEGARA constants with units"""

import astropy.units as u


# FIXME: duplicated in megaradrp.instrument
# without units

# Platescale in focal plane of Folded-Cass
GTC_FC_A_PLATESCALE = 1.212 * u.arcsec / u.mm

# Reference instrument aligment angle
MEGARA_IAA = -163.854 * u.deg

# mm from center to center, upwards
SPAXEL_SCALE = 0.443  * u.mm